@@ -1,7 +1,7 @@
 /*
     posix1e - a python module exposing the posix acl functions
 
-    Copyright (C) 2002-2008 Iustin Pop <iusty@k1024.org>
+    Copyright (C) 2002-2009 Iustin Pop <iusty@k1024.org>
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -32,7 +32,34 @@
 #define get_perm acl_get_perm_np
 #endif
 
-staticforward PyTypeObject ACL_Type;
+#if PY_MAJOR_VERSION >= 3
+#define IS_PY3K
+#define PyInt_Check(op) PyLong_Check(op)
+#define PyInt_FromString PyLong_FromString
+#define PyInt_FromUnicode PyLong_FromUnicode
+#define PyInt_FromLong PyLong_FromLong
+#define PyInt_FromSize_t PyLong_FromSize_t
+#define PyInt_FromSsize_t PyLong_FromSsize_t
+#define PyInt_AsLong PyLong_AsLong
+#define PyInt_AsSsize_t PyLong_AsSsize_t
+#define PyInt_AsUnsignedLongMask PyLong_AsUnsignedLongMask
+#define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
+#define PyInt_AS_LONG PyLong_AS_LONG
+#else
+#define PyBytes_Check PyString_Check
+#define PyBytes_AS_STRING PyString_AS_STRING
+#define PyBytes_FromStringAndSize PyString_FromStringAndSize
+#define PyBytes_FromString PyString_FromString
+#define PyBytes_FromFormat PyString_FromFormat
+#define PyBytes_ConcatAndDel PyString_ConcatAndDel
+
+/* Python 2.6 already defines Py_TYPE */
+#ifndef Py_TYPE
+#define Py_TYPE(o)    (((PyObject*)(o))->ob_type)
+#endif
+#endif
+
+static PyTypeObject ACL_Type;
 static PyObject* ACL_applyto(PyObject* obj, PyObject* args);
 static PyObject* ACL_valid(PyObject* obj, PyObject* args);
 
@@ -42,8 +69,8 @@ static PyObject* ACL_set_state(PyObject *obj, PyObject* args);
 #endif
 
 #ifdef HAVE_LEVEL2
-staticforward PyTypeObject Entry_Type;
-staticforward PyTypeObject Permset_Type;
+static PyTypeObject Entry_Type;
+static PyTypeObject Permset_Type;
 static PyObject* Permset_new(PyTypeObject* type, PyObject* args,
                              PyObject *keywds);
 #endif
@@ -99,11 +126,11 @@ static int ACL_init(PyObject* obj, PyObject* args, PyObject *keywds) {
 #ifdef HAVE_LINUX
     static char *kwlist[] = { "file", "fd", "text", "acl", "filedef",
                               "mode", NULL };
-    char *format = "|sisO!sH";
+    char *format = "|etisO!sH";
     mode_t mode = 0;
 #else
     static char *kwlist[] = { "file", "fd", "text", "acl", "filedef", NULL };
-    char *format = "|sisO!s";
+    char *format = "|etisO!s";
 #endif
     char *file = NULL;
     char *filedef = NULL;
@@ -118,7 +145,7 @@ static int ACL_init(PyObject* obj, PyObject* args, PyObject *keywds) {
         return -1;
     }
     if(!PyArg_ParseTupleAndKeywords(args, keywds, format, kwlist,
-                                    &file, &fd, &text, &ACL_Type,
+                                    NULL, &file, &fd, &text, &ACL_Type,
                                     &thesrc, &filedef
 #ifdef HAVE_LINUX
                                     , &mode
@@ -181,7 +208,7 @@ static PyObject* ACL_str(PyObject *obj) {
     if(text == NULL) {
         return PyErr_SetFromErrno(PyExc_IOError);
     }
-    ret = PyString_FromString(text);
+    ret = PyBytes_FromString(text);
     if(acl_free(text) != 0) {
         Py_DECREF(ret);
         return PyErr_SetFromErrno(PyExc_IOError);
@@ -235,7 +262,7 @@ static PyObject* ACL_to_any_text(PyObject *obj, PyObject *args,
     if(text == NULL) {
         return PyErr_SetFromErrno(PyExc_IOError);
     }
-    ret = PyString_FromString(text);
+    ret = PyBytes_FromString(text);
     if(acl_free(text) != 0) {
         Py_DECREF(ret);
         return PyErr_SetFromErrno(PyExc_IOError);
@@ -361,9 +388,18 @@ static PyObject* ACL_applyto(PyObject* obj, PyObject* args) {
     if (!PyArg_ParseTuple(args, "O|i", &myarg, &type))
         return NULL;
 
-    if(PyString_Check(myarg)) {
-        char *filename = PyString_AS_STRING(myarg);
+    if(PyBytes_Check(myarg)) {
+        char *filename = PyBytes_AS_STRING(myarg);
         nret = acl_set_file(filename, type, self->acl);
+    } else if (PyUnicode_Check(myarg)) {
+        PyObject *o =
+            PyUnicode_AsEncodedString(myarg,
+                                      Py_FileSystemDefaultEncoding, "strict");
+        if (o == NULL)
+            return NULL;
+        const char *filename = PyBytes_AS_STRING(o);
+        nret = acl_set_file(filename, type, self->acl);
+        Py_DECREF(o);
     } else if((fd = PyObject_AsFileDescriptor(myarg)) != -1) {
         nret = acl_set_fd(fd, self->acl);
     } else {
@@ -429,9 +465,9 @@ static PyObject* ACL_get_state(PyObject *obj, PyObject* args) {
     if(size == -1)
         return PyErr_SetFromErrno(PyExc_IOError);
 
-    if((ret = PyString_FromStringAndSize(NULL, size)) == NULL)
+    if((ret = PyBytes_FromStringAndSize(NULL, size)) == NULL)
         return NULL;
-    buf = PyString_AsString(ret);
+    buf = PyBytes_AsString(ret);
 
     if((nsize = acl_copy_ext(buf, self->acl, size)) == -1) {
         Py_DECREF(ret);
@@ -666,8 +702,7 @@ static PyObject* Entry_str(PyObject *obj) {
     acl_tag_t tag;
     uid_t qualifier;
     void *p;
-    PyObject *ret;
-    PyObject *format, *list;
+    PyObject *format, *kind;
     Entry_Object *self = (Entry_Object*) obj;
 
     if(acl_get_tag_type(self->entry, &tag) == -1) {
@@ -685,33 +720,31 @@ static PyObject* Entry_str(PyObject *obj) {
         qualifier = 0;
     }
 
-    format = PyString_FromString("ACL entry for %s");
+    format = PyBytes_FromString("ACL entry for ");
     if(format == NULL)
         return NULL;
-    list = PyTuple_New(1);
     if(tag == ACL_UNDEFINED_TAG) {
-        PyTuple_SetItem(list, 0, PyString_FromString("undefined type"));
+        kind = PyBytes_FromString("undefined type");
     } else if(tag == ACL_USER_OBJ) {
-        PyTuple_SetItem(list, 0, PyString_FromString("the owner"));
+        kind = PyBytes_FromString("the owner");
     } else if(tag == ACL_GROUP_OBJ) {
-        PyTuple_SetItem(list, 0, PyString_FromString("the group"));
+        kind = PyBytes_FromString("the group");
     } else if(tag == ACL_OTHER) {
-        PyTuple_SetItem(list, 0, PyString_FromString("the others"));
+        kind = PyBytes_FromString("the others");
     } else if(tag == ACL_USER) {
-        PyTuple_SetItem(list, 0, PyString_FromFormat("user with uid %d",
-                                                     qualifier));
+        kind = PyBytes_FromFormat("user with uid %d", qualifier);
     } else if(tag == ACL_GROUP) {
-        PyTuple_SetItem(list, 0, PyString_FromFormat("group with gid %d",
-                                                     qualifier));
+        kind = PyBytes_FromFormat("group with gid %d", qualifier);
     } else if(tag == ACL_MASK) {
-        PyTuple_SetItem(list, 0, PyString_FromString("the mask"));
+        kind = PyBytes_FromString("the mask");
     } else {
-        PyTuple_SetItem(list, 0, PyString_FromString("UNKNOWN_TAG_TYPE!"));
+        kind = PyBytes_FromString("UNKNOWN_TAG_TYPE!");
     }
-    ret = PyString_Format(format, list);
+    if (kind == NULL)
+        return NULL;
+    PyBytes_ConcatAndDel(&format, kind);
     Py_DECREF(format);
-    Py_DECREF(list);
-    return ret;
+    return format;
 }
 
 /* Sets the tag type of the entry */
@@ -934,7 +967,7 @@ static PyObject* Permset_str(PyObject *obj) {
     pstr[0] = get_perm(self->permset, ACL_READ) ? 'r' : '-';
     pstr[1] = get_perm(self->permset, ACL_WRITE) ? 'w' : '-';
     pstr[2] = get_perm(self->permset, ACL_EXECUTE) ? 'x' : '-';
-    return PyString_FromStringAndSize(pstr, 3);
+    return PyBytes_FromStringAndSize(pstr, 3);
 }
 
 static char __Permset_clear_doc__[] =
@@ -1132,8 +1165,12 @@ static PyMethodDef ACL_methods[] = {
 
 /* The definition of the ACL Type */
 static PyTypeObject ACL_Type = {
+#ifdef IS_PY3K
+    PyVarObject_HEAD_INIT(NULL, 0)
+#else
     PyObject_HEAD_INIT(NULL)
     0,
+#endif
     "posix1e.ACL",
     sizeof(ACL_Object),
     0,
@@ -1249,8 +1286,12 @@ static char __Entry_Type_doc__[] =
     ;
 /* The definition of the Entry Type */
 static PyTypeObject Entry_Type = {
+#ifdef IS_PY3K
+    PyVarObject_HEAD_INIT(NULL, 0)
+#else
     PyObject_HEAD_INIT(NULL)
     0,
+#endif
     "posix1e.Entry",
     sizeof(Entry_Object),
     0,
@@ -1354,8 +1395,12 @@ static char __Permset_Type_doc__[] =
 
 /* The definition of the Permset Type */
 static PyTypeObject Permset_Type = {
+#ifdef IS_PY3K
+    PyVarObject_HEAD_INIT(NULL, 0)
+#else
     PyObject_HEAD_INIT(NULL)
     0,
+#endif
     "posix1e.Permset",
     sizeof(Permset_Object),
     0,
@@ -1415,7 +1460,7 @@ static PyObject* aclmodule_delete_default(PyObject* obj, PyObject* args) {
     char *filename;
 
     /* Parse the arguments */
-    if (!PyArg_ParseTuple(args, "s", &filename))
+    if (!PyArg_ParseTuple(args, "et", NULL, &filename))
         return NULL;
 
     if(acl_delete_def_file(filename) == -1) {
@@ -1445,9 +1490,18 @@ static PyObject* aclmodule_has_extended(PyObject* obj, PyObject* args) {
     if (!PyArg_ParseTuple(args, "O", &myarg))
         return NULL;
 
-    if(PyString_Check(myarg)) {
-        const char *filename = PyString_AS_STRING(myarg);
+    if(PyBytes_Check(myarg)) {
+        const char *filename = PyBytes_AS_STRING(myarg);
         nret = acl_extended_file(filename);
+    } else if (PyUnicode_Check(myarg)) {
+        PyObject *o =
+            PyUnicode_AsEncodedString(myarg,
+                                      Py_FileSystemDefaultEncoding, "strict");
+        if (o == NULL)
+            return NULL;
+        const char *filename = PyBytes_AS_STRING(o);
+        nret = acl_extended_file(filename);
+        Py_DECREF(o);
     } else if((fd = PyObject_AsFileDescriptor(myarg)) != -1) {
         nret = acl_extended_fd(fd);
     } else {
@@ -1523,33 +1577,59 @@ static char __posix1e_doc__[] =
     "\n"
     ;
 
-void initposix1e(void) {
+#ifdef IS_PY3K
+
+static struct PyModuleDef posix1emodule = {
+    PyModuleDef_HEAD_INIT,
+    "posix1e",
+    __posix1e_doc__,
+    0,
+    aclmodule_methods,
+};
+
+#define INITERROR return NULL
+
+PyMODINIT_FUNC
+PyInit_posix1e(void)
+
+#else
+#define INITERROR return
+
+void initposix1e(void)
+#endif
+{
     PyObject *m, *d;
 
-    ACL_Type.ob_type = &PyType_Type;
+    Py_TYPE(&ACL_Type) = &PyType_Type;
     if(PyType_Ready(&ACL_Type) < 0)
-        return;
+        INITERROR;
 
 #ifdef HAVE_LEVEL2
-    Entry_Type.ob_type = &PyType_Type;
+    Py_TYPE(&Entry_Type) = &PyType_Type;
     if(PyType_Ready(&Entry_Type) < 0)
-        return;
+        INITERROR;
 
-    Permset_Type.ob_type = &PyType_Type;
+    Py_TYPE(&Permset_Type) = &PyType_Type;
     if(PyType_Ready(&Permset_Type) < 0)
-        return;
+        INITERROR;
 #endif
 
+#ifdef IS_PY3K
+    m = PyModule_Create(&posix1emodule);
+#else
     m = Py_InitModule3("posix1e", aclmodule_methods, __posix1e_doc__);
+#endif
+    if (m==NULL)
+        INITERROR;
 
     d = PyModule_GetDict(m);
     if (d == NULL)
-        return;
+        INITERROR;
 
     Py_INCREF(&ACL_Type);
     if (PyDict_SetItemString(d, "ACL",
                              (PyObject *) &ACL_Type) < 0)
-        return;
+        INITERROR;
 
     /* 23.3.6 acl_type_t values */
     PyModule_AddIntConstant(m, "ACL_TYPE_ACCESS", ACL_TYPE_ACCESS);
@@ -1560,12 +1640,12 @@ void initposix1e(void) {
     Py_INCREF(&Entry_Type);
     if (PyDict_SetItemString(d, "Entry",
                              (PyObject *) &Entry_Type) < 0)
-        return;
+        INITERROR;
 
     Py_INCREF(&Permset_Type);
     if (PyDict_SetItemString(d, "Permset",
                              (PyObject *) &Permset_Type) < 0)
-        return;
+        INITERROR;
 
     /* 23.2.2 acl_perm_t values */
     PyModule_AddIntConstant(m, "ACL_READ", ACL_READ);
@@ -1612,4 +1692,8 @@ void initposix1e(void) {
     PyModule_AddIntConstant(m, "HAS_EXTENDED_CHECK", 0);
     PyModule_AddIntConstant(m, "HAS_EQUIV_MODE", 0);
 #endif
+
+#ifdef IS_PY3K
+    return m;
+#endif
 }
